# ##### BEGIN GPL LICENSE BLOCK #####
#
# This file is part of the BlenderAe Addon.
#
# BlenderAe is a Blender Addon that allows exporting of Blender data for After Effects.
# Copyright (C) 2021 Daniel Peterson
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of  MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ##### END GPL LICENSE BLOCK #####

import json
import pathlib
import subprocess
import sys
from inspect import getsourcefile
from math import degrees, floor, radians
from mathutils import Euler, Matrix
from numpy import radians

import bpy

# from BlenderAe_Export import find_ae_path


def load_json(f):
    try:
        with open(f, encoding="utf-8") as file:
            print("Loaded data file")
            return json.load(file)

    except Exception as e:
        print(f"Exception: {e}")
        return False


def ae_loc_conv(
    ae_loc, aspect, width, height, scale_factor, solid_child=False, child=False
):
    """Convert location to Blender coordinates"""
    # BUG - For solids (and maybe other types of layers) the children layer relative positions
    # are calculated from the top left corner [0,0], not the anchor point location.

    if bpy.context.scene.blenderae.center_offset and not child and not solid_child:
        ae_x = ae_loc[0] - width / 2
        ae_y = ae_loc[1] - height / 2

    elif solid_child:
        ae_x = ae_loc[0] - width / 2
        ae_y = ae_loc[1] - height / 2

    else:
        ae_x = ae_loc[0]
        ae_y = ae_loc[1]

    ae_z = ae_loc[2]

    # # From Ae to Blender x,y,z
    bl_x = ae_x * aspect * scale_factor  # calculate Blender X position
    bl_y = ae_z * scale_factor  # calculate Blender Z position
    bl_z = -ae_y * scale_factor  # calculate Blender Y position

    return [bl_x, bl_y, bl_z]


def ae_rot_conv(ae_rot, local=False):
    # # Convert rotations assuming 'ZYX' euler mode.

    ae_rx = ae_rot[0]
    ae_ry = ae_rot[1]
    ae_rz = ae_rot[2]

    bl_rx = radians(ae_rx) if local else radians(ae_rx + 90)
    bl_ry = -radians(ae_ry)  # AE's Y orientation == negative Blender Y rotation
    bl_rz = -radians(ae_rz)  # AE's Z orientation == negative Blender Z rotation
    return [bl_rx, bl_ry, bl_rz]


# Local rotation functions
# Set local rotation on x,y,z
def rotate_local(obj, global_rotation, local_rotation):
    """rotates an object around a local x,y,z rotation"""

    local_rot_euler = Euler(local_rotation, "ZYX")
    # print(f"Local rotation: {local_rot_euler}")

    obj.rotation_euler = (
        global_rotation.to_matrix() @ local_rot_euler.to_matrix()
    ).to_euler(obj.rotation_mode)
    # print(f"New object Euler with local values: {obj.rotation_euler} \n")
    return


# # set local rotation by single axis
def rotate_local_single_axis(obj, angle_degrees, axis="Z"):
    """rotates an object around a local single axis"""
    # local rotation about axis
    obj.rotation_euler = (
        obj.rotation_euler.to_matrix()
        @ Matrix.Rotation(radians(angle_degrees), 3, axis)
    ).to_euler()


def ae_scale_conv(ae_scale):
    # Blender scale is 1.0, needs to be 100% in AE
    # For a plane x,y scale are the same as Ae x,y (plane lies horizontal whereas in Ae it's vertical BUT x,y are the same 'relative' to the plane face).

    bl_sx = ae_scale[0] / 100
    bl_sy = ae_scale[1] / 100
    bl_sz = ae_scale[2] / 100
    return [bl_sx, bl_sy, bl_sz]


def ae_lens(zoom, width, sensor, aspect):
    # Blender to Ae zoom conversion
    # zoom = camera.data.lens * dimension / sensor * aspect

    # eg.
    # camera.data.lens = 50mm
    # dimension = 1920
    # sensor = 36mm
    # aspect = 1

    # eg. from ae data
    # zoom = 1765.8183760382
    # dimension = 1920
    # sensor = 36 (default - non readable in ae scripting)
    # aspect = 1
    # 1765.8183760382 = lens * 1920/36 * 1
    # lens = 1765.8183760382 * 36 / 1920 / 1
    # lens = 33.11 (correct in ae)

    lens = zoom * sensor / width / aspect

    return lens


# def file_checker(f):
#     # Check for the file, sleep if not
#     iteration = 0
#     max_iteration = 6
#     time_to_sleep = 1 #seconds
#     file_present = False

#     # print ("File checker: ", f)
#     # print ("File type: ", type(f))

#     while (not file_present) and (iteration < max_iteration):
#         if pathlib.Path.is_file(f):
#             print(f"{f} ...File exists")
#             file_present = True
#             break

#         else:
#             print(f"Waiting for data... {iteration*time_to_sleep} seconds")
#             time.sleep(time_to_sleep)
#             iteration += 1

#     return file_present


def get_data_folder_with_file(file_name):
    # load directory and load json data if available
    directory = pathlib.Path(__file__).parent.resolve()
    data_folder = directory / "data"

    # check for a data folder, if it doesn't exist create it
    if not pathlib.Path.is_dir(data_folder):
        pathlib.Path.mkdir(data_folder)

    return data_folder / file_name

def get_root_with_file(file_name):
    # load directory and load script file if available
    directory = pathlib.Path(__file__).parent.resolve()
    return directory / file_name


def create_empty(name: str, size=1, display_type="ARROWS"):
    """create an empty with a name and type"""

    empty_obj = bpy.data.objects.new(name, None)
    empty_obj.empty_display_size = size

    # enum in ['PLAIN_AXES', 'ARROWS', 'SINGLE_ARROW', 'CIRCLE', 'CUBE', 'SPHERE', 'CONE', 'IMAGE']
    empty_obj.empty_display_type = display_type

    # ZYX matches Ae rotation space
    empty_obj.rotation_mode = "ZYX"

    return empty_obj


def send_to_ae(ae_path, script_path):

    if ae_path != "" and script_path != "":
        # Windows
        if sys.platform == "win32":
            cmd = ae_path + " -r " + script_path
            # print ("cmd: ", cmd)
            try:
                send_data = subprocess.run(
                    cmd, text=True, universal_newlines=True, timeout=4
                )

                if send_data.stderr:
                    print(f"Error sending jsx script to Ae: {send_data.stderr}")
                    return False
                else:
                    print("Layers import script sent to Ae")
                    return True

            except subprocess.TimeoutExpired:
                print("Layers import subprocess timeout")
                # Assume still succesfully sent, true
                return True

        # Mac
        elif sys.platform == "darwin":
            cmd = f"""
                tell application "{ae_path}"
                    DoScriptFile "{script_path}"
                end tell
                """
            send_data = subprocess.run(
                ["osascript", "-e", cmd], capture_output=True, text=True
            )

            # print("stderr: ", send_data.stderr)
            # print("stdout: ", send_data.stdout)
            # print("return: ", send_data.returncode)

            if send_data.returncode == 0:
                print("Layers import script sent to Ae")
                return True

            else:
                print("Layers import subprocess error")
                return False
                # If no path is set copy the import script to the Documents\Blender_Ae folder

        else:
            print("Failed - script_path() or ae_path() was empty")
            return False

    else:
        print("System not detected as win32 or MacOS(darwin)!")
        return False


def ae_subprocess_execute():
    # Check for previous json file and if true delete the old file
    json_name = "BlenderAe_AeLayers.json"
    json_path = get_data_folder_with_file(json_name)

    if pathlib.Path.is_file(json_path):
        json_path.unlink(missing_ok=True)

    # Get active Ae path
    ae_path = bpy.context.scene.blenderae.ae_filepath
    print("......\nRunning BlenderAe Importer\n")

    # Get script path
    script_name = "BlenderAe_AeLayers.jsxbin"
    script_path = get_root_with_file(script_name)
    if pathlib.Path.is_file(script_path):
        # Subprocess sends script to the active After effects. Only run after checking the script path is valid.
        return bool(send_to_ae(str(ae_path), str(script_path)))

    # If previous script name is not a file check for jsx instead
    script_name = "BlenderAe_AeLayers.jsx"
    script_path = get_root_with_file(script_name)
    if pathlib.Path.is_file(script_path):
        return bool(send_to_ae(str(ae_path), str(script_path)))

    print("Script file doesn't exist")
    return False


def animate_null(data, curr_scene, width, height, aspect, scale_factor, key):

    null_obj = data[key]["obj"]
    parent_check = data[key]["parent"]
    parent_solid_check = False

    if data[key]["parentType"] == "Solid":
        width, height = solidParentDimensions(data, key)
        parent_solid_check = True

    # Properties
    location = data[key]["nullPosition"]
    rotation = data[key]["nullOrientation"]
    scale = data[key]["nullScale"]
    local_rot = data[key]["localRotationValues"]

    # Flags for case of single keyframes
    loc_skip = False
    rot_skip = False
    local_rot_skip = False
    scale_skip = False

    for i, f in enumerate(
        range(data[key]["nullInPoint"] + 1, data[key]["nullOutPoint"] + 1)
    ):
        curr_scene.frame_set(f)

        # Location
        if not loc_skip:
            if len(location) == 1:
                null_obj.location = ae_loc_conv(
                    location[0],
                    aspect,
                    width,
                    height,
                    scale_factor,
                    parent_solid_check,
                    parent_check,
                )
                loc_skip = True

            else:
                null_obj.location = ae_loc_conv(
                    location[i],
                    aspect,
                    width,
                    height,
                    scale_factor,
                    parent_solid_check,
                    parent_check,
                )
                null_obj.keyframe_insert("location")

        # Rotation
        if not rot_skip:
            if len(rotation) == 1:
                null_obj.rotation_euler = ae_rot_conv(rotation[0])
                rot_skip = True

            else:
                null_obj.rotation_euler = ae_rot_conv(rotation[i])
                null_obj.keyframe_insert("rotation_euler")

            # Need to set an initial orientation value for the local rotation matrix multiplication...
            # Otherwise it results in exponentially increasing values based on the previous frames null_obj.rotation_euler (which is the local rotation)...
            # Only an issue when there's no keyframes (1 value) on the orientation (global rotation).
            # print (f"global rotation inside: {global_rotation}")
            global_rotation = null_obj.rotation_euler.copy()

        # If local rotation is set then we need to update keyframe with local rotation data for each keyframe
        if local_rot:
            if len(local_rot) == 1:
                local_rot_conv = ae_rot_conv(local_rot[0], local=True)

            else:
                local_rot_conv = ae_rot_conv(local_rot[i], local=True)

            # print (f"global rotation: {global_rotation}")
            # print (f"Initial null_obj.rotation_euler: {null_obj.rotation_euler}")
            rotate_local(null_obj, global_rotation, local_rot_conv)
            null_obj.keyframe_insert("rotation_euler")

        # Scale
        if not scale_skip:
            if len(scale) == 1:
                null_obj.scale = ae_scale_conv(scale[0])
                scale_skip = True

            else:
                null_obj.scale = ae_scale_conv(scale[i])
                null_obj.keyframe_insert("scale")

    bpy.context.view_layer.update()


def solidParentDimensions(data, key):
    """Get the solid dimensions from the parent to offset the obj position"""
    # loop through data keys and find the matching parent by index (names can be duplicated)
    for parent_key in data:
        if (
            "index" in data[parent_key].keys()
            and data[parent_key]["index"] == data[key]["parentIndex"]
        ):
            # get the solid width and height. Replace the comp width and height variables
            width = data[parent_key]["solidWidth"]
            height = data[parent_key]["solidHeight"]

            return width, height


def create_null(data_key, ae_collection):
    # Create empty
    null_name = data_key["name"]
    null_obj = bpy.data.objects.new(null_name, None)
    null_obj.empty_display_size = 1
    null_obj.empty_display_type = "ARROWS"

    # ZYX matches Ae rotation space
    null_obj.rotation_mode = "ZYX"

    # Add empty to scene collection
    ae_collection.objects.link(null_obj)

    return null_obj


def animate_solid(data, curr_scene, width, height, aspect, scale_factor, key):

    plane_obj = data[key]["obj"]
    parent_check = data[key]["parent"]
    parent_solid_check = False

    # The plane position only needs the width and height subtracted if the parent is a solid, adjustment.
    # This is because a planes position in Blender is at the center whereas in After Effects it is at the top left.
    # ie. The top left from the center is -width/2 and -height/2
    # Check if the obj has a parent that is a solid.
    if data[key]["parentType"] in ["Solid"]:
        width, height = solidParentDimensions(data, key)
        parent_solid_check = True

    # Properties
    location = data[key]["solidPosition"]
    rotation = data[key]["solidOrientation"]
    scale = data[key]["solidScale"]
    local_rot = data[key]["localRotationValues"]

    # Flags for case of no keyframes (single value)
    loc_skip = False
    rot_skip = False
    scale_skip = False

    for i, f in enumerate(
        range(data[key]["solidInPoint"] + 1, data[key]["solidOutPoint"] + 1)
    ):
        curr_scene.frame_set(f)

        # Location
        if not loc_skip:
            if len(location) == 1:
                plane_obj.location = ae_loc_conv(
                    location[0],
                    aspect,
                    width,
                    height,
                    scale_factor,
                    parent_solid_check,
                    parent_check,
                )
                loc_skip = True

            else:
                plane_obj.location = ae_loc_conv(
                    location[i],
                    aspect,
                    width,
                    height,
                    scale_factor,
                    parent_solid_check,
                    parent_check,
                )
                plane_obj.keyframe_insert("location")

        # Rotation Global
        if not rot_skip:
            if len(rotation) == 1:
                plane_obj.rotation_euler = ae_rot_conv(rotation[0])
                rot_skip = True

            else:
                plane_obj.rotation_euler = ae_rot_conv(rotation[i])
                plane_obj.keyframe_insert("rotation_euler")

            global_rotation = plane_obj.rotation_euler.copy()

        # Rotation Local
        if local_rot:
            if len(local_rot) == 1:
                local_rot_conv = ae_rot_conv(local_rot[0], local=True)
            else:
                local_rot_conv = ae_rot_conv(local_rot[i], local=True)

            rotate_local(plane_obj, global_rotation, local_rot_conv)
            plane_obj.keyframe_insert("rotation_euler")

        # Scale - Dimensions of solids is multiplied by the scale factor so it's not needed for the actual scale, just normal conversion.
        if not scale_skip:
            if len(scale) == 1:
                plane_obj.scale = ae_scale_conv(scale[0])
                scale_skip = True

            else:
                plane_obj.scale = ae_scale_conv(scale[i])
                plane_obj.keyframe_insert("scale")

        bpy.context.view_layer.update()


def create_solid(data_key, scale_factor, ae_collection):
    # bpy.ops method ... https://docs.blender.org/api/current/bpy.ops.mesh.html#bpy.ops.mesh.primitive_plane_add
    # bpy.ops.mesh.primitive_plane_add(size=2.0, calc_uvs=True, enter_editmode=False, align='WORLD', location=(0.0, 0.0, 0.0), rotation=(0.0, 0.0, 0.0), scale=(0.0, 0.0, 0.0))

    # Manual methods
    # https://b3d.interplanety.org/en/how-to-create-mesh-through-the-blender-python-api/
    # https://blenderartists.org/t/adding-a-plane-without-using-bpy-ops/1248056/3

    solid_width = data_key["solidWidth"] / 2 * scale_factor
    solid_height = data_key["solidHeight"] / 2 * scale_factor
    solid_colour = data_key["solidColour"]
    solid_name = data_key["name"]

    x = solid_width
    y = solid_height

    vert = [(-x, -y, 0.0), (x, -y, 0.0), (-x, y, 0.0), (x, y, 0.0)]
    edge = []
    fac = [(0, 1, 3, 2)]

    plane_data = bpy.data.meshes.new(solid_name)
    plane_data.from_pydata(vert, edge, fac)
    plane_obj = bpy.data.objects.new(solid_name, plane_data)

    # ZYX matches Ae rotation space
    plane_obj.rotation_mode = "ZYX"

    # add object to scene collection
    ae_collection.objects.link(plane_obj)

    return plane_obj


def create_camera_rig(data_key, ae_collection):
    cam = bpy.data.cameras.new(data_key["name"])
    cam_obj = bpy.data.objects.new(data_key["name"], cam)

    # ZYX matches Ae rotation space
    cam_obj.rotation_mode = "ZYX"

    # add object to scene collection
    ae_collection.objects.link(cam_obj)

    poi = data_key["camPOI"]

    # Point of Interest
    if poi:
        poi_name = data_key["name"] + "_POI"
        poi_empty = create_empty(poi_name, size=0.5, display_type="ARROWS")
        poi_empty.rotation_mode = "ZYX"

        rot_name = data_key["name"] + "_Track_to_POI"
        track_to_empty = create_empty(rot_name, size=0.5, display_type="ARROWS")
        track_to_empty.rotation_mode = "ZYX"

        # Add empty to scene collection
        ae_collection.objects.link(poi_empty)
        ae_collection.objects.link(track_to_empty)

        # parenting
        cam_obj.parent = track_to_empty

        # constraint = cam_obj.constraints.new(type='DAMPED_TRACK')
        # Ae Cameras POI match TRACK_TO constraint
        constraint = track_to_empty.constraints.new(type="TRACK_TO")
        constraint.target = poi_empty
        constraint.track_axis = "TRACK_NEGATIVE_Z"
        constraint.use_target_z = True

    else:
        poi_empty = False
        track_to_empty = False

    return cam_obj, poi_empty, track_to_empty


def animate_camera_rig(
    data, curr_scene, width, height, sensor, aspect, scale_factor, key
):

    cam_obj = data[key]["obj"]
    poi_empty = data[key]["poi_empty"]
    track_to_empty = data[key]["track_to_empty"]

    # Note that the camera object name is different to the camera data name. 
    # Important to use the cam_obj.data.name to get the correct camera here.
    cam_data = bpy.data.cameras[cam_obj.data.name]

    # Set the target obj for the location animation. If the parent has it's own parent then don't change the values
    target_obj = track_to_empty or cam_obj
    parent_check = data[key]["parent"]
    parent_solid_check = False

    # Check if the cam_obj has a parent that is a solid.
    # Replace the width and height with the solids dictionary width and height.
    # (Because the parented object is in relation to the top left of the solid, not the comp dimensions)
    if data[key]["parentType"] == "Solid":
        width, height = solidParentDimensions(data, key)
        parent_solid_check = True

    # cam_obj properties
    zoom = data[key]["camZoom"]
    location = data[key]["camPosition"]
    rotation = data[key]["camOrientation"]
    poi = data[key]["camPOI"]
    local_rot = data[key]["localRotationValues"]

    # Flags for case of 0 keyframes
    zoom_skip = False
    loc_skip = False
    rot_skip = False
    poi_skip = False

    for i, f in enumerate(
        range(data[key]["camInPoint"] + 1, data[key]["camOutPoint"] + 1)
    ):
        curr_scene.frame_set(f)

        # Zoom
        # print("zoom_skip", zoom_skip)

        if not zoom_skip:
            if len(zoom) == 1:
                cam_data.lens = ae_lens(zoom[0], width, sensor, aspect)
                zoom_skip = True
                # print("cam_data.lens", cam_data.lens)

            else:
                # Lens animation is assigned to a camera data not 'cam_obj' object type.
                # https://blender.stackexchange.com/questions/87680/animating-cameras-focal-length-via-python-code
                cam_data.lens = ae_lens(zoom[i], width, sensor, aspect)
                cam_data.keyframe_insert("lens")

        # Location
        if not loc_skip:
            if len(location) == 1:
                target_obj.location = ae_loc_conv(
                    location[0],
                    aspect,
                    width,
                    height,
                    scale_factor,
                    parent_solid_check,
                    parent_check,
                )
                loc_skip = True

            else:
                target_obj.location = ae_loc_conv(
                    location[i],
                    aspect,
                    width,
                    height,
                    scale_factor,
                    parent_solid_check,
                    parent_check,
                )
                target_obj.keyframe_insert("location")

        # track_to_empty only tracks to the target point of interest if it exists, doesn't need rotation values.
        # Also the POI emtpy get's parented to objects and the track_to_empty copies the
        if not rot_skip:
            # check if POI exists, rotation doesn't need the +90 on the x rot because of the parenting
            local_check = bool(poi)

            # Rotation Global
            if len(rotation) == 1:
                cam_obj.rotation_euler = ae_rot_conv(rotation[0], local=local_check)
                rot_skip = True

            else:
                cam_obj.rotation_euler = ae_rot_conv(rotation[i], local=local_check)
                cam_obj.keyframe_insert("rotation_euler")

            global_rotation = cam_obj.rotation_euler.copy()

        # Rotation Local
        if local_rot:
            if len(local_rot) == 1:
                local_rot_conv = ae_rot_conv(local_rot[0], local=True)
            else:
                local_rot_conv = ae_rot_conv(local_rot[i], local=True)

            rotate_local(cam_obj, global_rotation, local_rot_conv)
            cam_obj.keyframe_insert("rotation_euler")

        # Point of Interest
        if not poi_skip and poi:
            if len(poi) == 1:
                # no keyframes
                poi_empty.location = ae_loc_conv(
                    poi[0],
                    aspect,
                    width,
                    height,
                    scale_factor,
                    parent_solid_check,
                    parent_check,
                )
                poi_skip = True

            else:
                # Do animation keyframe inserts on the empty
                poi_empty.location = ae_loc_conv(
                    poi[i],
                    aspect,
                    width,
                    height,
                    scale_factor,
                    parent_solid_check,
                    parent_check,
                )
                poi_empty.keyframe_insert("location")

    bpy.context.view_layer.update()


def clean_keyframes_for_object(ob):

    curr_selection = bpy.context.selected_objects
    bpy.ops.object.select_all(action="DESELECT")

    ob.select_set(True)

    for area in bpy.context.screen.areas:
        if area.type == "GRAPH_EDITOR":
            override = bpy.context.copy()
            override["area"] = area
            try:
                bpy.ops.graph.select_all(override, action="SELECT")
                bpy.ops.graph.clean(override, channels=True)

            except Exception as e:
                print(f"Clean error or no keyframes: {e}")

    bpy.ops.object.select_all(action="DESELECT")
    for ob in curr_selection:
        ob.select_set(True)


# def clean_keyframes_for_selected_switch(obs):

#     curr_selection = bpy.context.selected_objects
#     bpy.ops.object.select_all(action='DESELECT')

#     for ob in obs:
#         ob.select_set(True)

#     area = bpy.context.area
#     print(f"area: {area}")
#     old_type = area.type
#     area.type = 'GRAPH_EDITOR'
#     bpy.ops.graph.select_all(action='SELECT')
#     bpy.ops.graph.clean(channels=True)
#     area.type = old_type

#     bpy.ops.object.select_all(action='DESELECT')
#     for ob in curr_selection:
#         ob.select_set(True)


def create_and_animate(data):

    def remove_unused_camera_data():
        # Create a list to store the names of unused camera data
        unused_camera_data_names = []

        # Iterate through all camera data in bpy.data.cameras
        for camera_data in bpy.data.cameras:
            # Check if the camera data has any linked camera objects
            if not camera_data.users:
                unused_camera_data_names.append(camera_data.name)

        # Remove the unused camera data
        for data_name in unused_camera_data_names:
            bpy.data.cameras.remove(bpy.data.cameras[data_name])

    # Remove unused camera data that may be present from previous imports from Ae
    remove_unused_camera_data()

    if data:
        curr_scene = bpy.context.scene
        # comp_info = list(data.keys())[0]
        comp_settings = data["comp_settings"]
        width = comp_settings["width"]
        height = comp_settings["height"]

        # In Ae changing the comp resolution changes the appearance/scale of layers in relation to the comp.
        # An Ae camera focal length also changes to match changed comp dimensions.
        # In Blender the resolution is independent to the object scale (much better), only the aspect ratio of the camera changes if needed.
        # Currently sensor is set to ae default - not able to obtain this value in extendscript.
        sensor = 36
        aspect = comp_settings["aspectRatio"]

        # Scale factor, 0.001 is one pixel = one millimeter with default Blender settings of 1unit = 1m
        # Note that Ae camera tracking uses abritrary units that don't match real world scale.
        scale_factor = bpy.context.scene.blenderae.scale_factor / 1000

        # check or make collection
        ae_collection = bpy.data.collections.get("ae_collection")
        if ae_collection is None:
            ae_collection = bpy.data.collections.new("ae_collection")

        if "ae_collection" not in curr_scene.collection.children:
            curr_scene.collection.children.link(ae_collection)

        # loop through keys in data to create all objects
        # Need to create all objects first to create parent links
        for obj_key in data:
            # create objects
            if data[obj_key]["type"] == "Camera":
                cam_obj, poi_empty, track_to_empty = create_camera_rig(
                    data[obj_key], ae_collection
                )
                # Add objects to dict
                data[obj_key]["obj"] = cam_obj
                data[obj_key]["poi_empty"] = poi_empty
                data[obj_key]["track_to_empty"] = track_to_empty

            elif data[obj_key]["type"] in ["Solid", "Adjustment", "Footage", "Shape", "Precomp"]:
                obj = create_solid(data[obj_key], scale_factor, ae_collection)
                data[obj_key]["obj"] = obj
                # Convert initial rotation (will be recalculated in animation)

            elif data[obj_key]["type"] == "Null":
                obj = create_null(data[obj_key], ae_collection)
                data[obj_key]["obj"] = obj

        # Parent objects
        for obj_key in data:
            # if there is a parent key and it is not false
            if "parent" in data[obj_key].keys() and data[obj_key]["parent"]:

                # Loop through again to get the parent data
                for parent_obj_key in data:
                    # If the obj index equals the parent_index, set that obj to be the parent
                    if ("index" in data[parent_obj_key].keys()) and (
                        data[parent_obj_key]["index"] == data[obj_key]["parentIndex"]
                    ):

                        parent_obj = data[parent_obj_key]["obj"]
                        child_obj = data[obj_key]["obj"]

                        # If the parent is a solid, set dictionary key to True
                        data[obj_key]["parent_solid"] = (
                            parent_obj.type == "MESH"
                            and data[parent_obj_key]["type"] == "Solid"
                        )

                        # Check if the child already has a parent, set the child_obj to the top most parent
                        while child_obj.parent is not None:
                            child_obj = child_obj.parent
                        print(f"{child_obj.name} parent: {parent_obj.name}")

                        # Need the parents orientation to be initially converted to factor in +90 degrees on x.
                        parent_obj.rotation_euler = ae_rot_conv([0, 0, 0], local=False)

                        # update the depsgraph
                        depsgraph = bpy.context.evaluated_depsgraph_get()
                        parent_obj_eval = parent_obj.evaluated_get(depsgraph)

                        # Keep transformation difference (just the rotation difference currently as locations haven't been animated yet).
                        child_obj.parent = parent_obj
                        child_obj.matrix_parent_inverse = (
                            parent_obj_eval.matrix_world.inverted()
                        )

                        # For cameras with a poi it is not parented to anything... parent it to the main parent_obj
                        if (
                            "poi_empty" in data[obj_key].keys()
                            and data[obj_key]["poi_empty"]
                        ):

                            poi_empty = data[obj_key]["poi_empty"]
                            poi_empty.parent = parent_obj
                            poi_empty.matrix_parent_inverse = (
                                parent_obj_eval.matrix_world.inverted()
                            )

        # Animate objects.
        # Only need to convert rotations for the parents.
        for obj_key in data:
            if data[obj_key]["type"] == "Camera":
                animate_camera_rig(
                    data,
                    curr_scene,
                    width,
                    height,
                    sensor,
                    aspect,
                    scale_factor,
                    obj_key,
                )

            elif data[obj_key]["type"] in ["Solid","Adjustment","Footage","Shape","Precomp"]:
                animate_solid(
                    data,
                    curr_scene,
                    width,
                    height,
                    aspect,
                    scale_factor,
                    obj_key,
                )

            elif data[obj_key]["type"] == "Null":

                animate_null(
                    data,
                    curr_scene,
                    width,
                    height,
                    aspect,
                    scale_factor,
                    obj_key,
                )

            if "obj" in data[obj_key].keys() and data[obj_key]["obj"]:
                clean_keyframes_for_object(data[obj_key]["obj"])

        return True

    else:
        print("Import Json data file not found...")
        return False


def data_import():

    json_name = "BlenderAe_AeLayers.json"
    json_path = get_data_folder_with_file(json_name)

    iteration = 0
    def timer():
        nonlocal iteration
        max_iteration = 15
        time_to_sleep = 1.0  # seconds

        print(f"Waiting for data... {iteration*time_to_sleep} seconds")

        iteration += 1
        if pathlib.Path.is_file(json_path):

            # Load json data and create scene
            print(f"Ae layers data found: {json_path}")

            data = load_json(json_path)
            create_and_animate(data)

            # Delete json file otherwise the importer uses the left over data before waiting for the new file
            # BUG - This doesn't remove the file here for some reason.
            # json_path.unlink(missing_ok=True)
            return None

        elif iteration == max_iteration:
            print("No file found")
            return None

        return time_to_sleep

        # Check for the json file through application timer

    try:
        bpy.app.timers.register(timer)
    except Exception as e:
        print(e)
