# ##### BEGIN GPL LICENSE BLOCK #####
#
# This file is part of the BlenderAe Addon.
#
# BlenderAe is a Blender Addon that allows exporting of Blender data for After Effects.
# Copyright (C) 2021 Daniel Peterson
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of  MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ##### END GPL LICENSE BLOCK #####

# import datetime
# import time

import json
import os
import platform
import shutil
import subprocess
import sys
from inspect import getsourcefile
from math import degrees, floor
from pathlib import Path

try:
    # import psutil from the dependencies folder using pathlib
    addon_path = Path(__file__).parent.absolute()
    dependencies_path = addon_path / "dependencies"
    sys.path.append(str(dependencies_path))

    import psutil

except ImportError:
    print(
        "psutil import error, please install the BlenderAe dependencies in the addon preferences and restart Blender."
    )

import bmesh
import bpy
import mathutils
import numpy as np


# ----------
# Functions
# ----------
# def getBlenderPath():
#     blender_path = os.path.dirname(sys.executable)
#     file_name = bpy.path.basename(bpy.context.blend_data.filepath)
#     blender_file_info = [blender_path, file_name]
#     return blender_file_info
def clean_name(ob) -> str:
    """
    clean path or object by only keeping alphanumeric and replacing first digit character.
    https://docs.python.org/3/library/re.html#re.sub
    https://stackoverflow.com/questions/27938765/replace-non-alphanumeric-characters-except-some-exceptions-python
    """
    # Remove numbers at the begining of a name (for After Effects variables).
    if ob.name[0].isdigit():
        return f"_{bpy.path.clean_name(ob.name)}"
    else:
        return bpy.path.clean_name(ob.name)


def get_blender_path() -> list:
    """
    clean filename and return list
    """
    blender_path = bpy.data.filepath
    file_name = os.path.splitext(bpy.path.basename(bpy.data.filepath))[0]
    file_name_clean = bpy.path.clean_name(file_name)

    return [blender_path, file_name_clean]


# Create dictionary of selected items
def get_selected_objects(obs):

    cameras = []
    lights = []
    nulls = []
    faces = []  # edit mode
    verts = []  # edit mode
    bones = []  # object, edit or pose mode
    # text = [] #WIP

    for ob in obs:
        new_name = clean_name(ob)
        # check if object is a mesh
        if ob.type == "ARMATURE":
            # append bones to list
            bones.extend(
                [ob, new_name, bone]
                for bone in ob.pose.bones
                if bone.bone.select
            )
        elif bpy.context.object.mode == "EDIT":
            bm = bmesh.new()
            bm = bmesh.from_edit_mesh(ob.data).copy()
            bm.select_flush(True)

            # print(tuple(bpy.context.tool_settings.mesh_select_mode))

            # check if any faces are selected in faces mode
            if (bpy.context.tool_settings.mesh_select_mode[2]) and (
                True in [f.select for f in bm.faces]
            ):
                print("Faces are Selected")
                for f in bm.faces:
                    if f.select:
                        faces.append([ob, new_name])
                        break

            elif (bpy.context.tool_settings.mesh_select_mode[0]) and (
                True in [v.select for v in bm.verts]
            ):  # check if any verts are selected in vertex mode
                print("verts selected: ", [x.select for x in bm.verts])
                for v in bm.verts:
                    if v.select:
                        verts.append([ob, new_name])
                        break

            bm.free()

        elif bpy.context.object.mode == "OBJECT":
            if ob.type == "CAMERA":
                cameras.append([ob, new_name])

            elif ob.type == "LIGHT":
                lights.append([ob, f"{new_name}_{ob.data.type}"])

            elif check_planar(ob) == True:
                faces.append([ob, new_name])
                # solids.append([ob, new_name])

            else:
                nulls.append([ob, new_name])

    selection = {
        "cameras": cameras,
        "lights": lights,
        "nulls": nulls,
        "faces": faces,
        "verts": verts,
        "bones": bones,
        #'text': text
    }

    print("Selection: ", selection)
    return selection


# Convert world matrix transforms to Ae transforms
def Ae_transforms(context):
    loc = list(context.matrix_world.to_translation())
    rot = list(context.matrix_world.to_euler("ZYX"))  # Ae's rotation space
    scale = list(context.matrix_world.to_scale())
    # print (f"Ae transform scale {scale}")

    x, y, z = loc_conv(loc)
    rx, ry, rz = rot_conv(rot)
    sx, sy, sz = scale_conv(scale)

    return x, y, z, rx, ry, rz, sx, sy, sz

    # loc = context.matrix_world.to_translation()
    # rot = context.matrix_world.to_euler('ZYX') #Ae's rotation space
    # scale = context.matrix_world.to_scale()


def loc_conv(loc, center_offset=True):
    if len(loc) == 3:
        # print (f"loc: {loc}")
        bx = loc[0]
        by = loc[1]
        bz = loc[2]

    # For calculating 2d face shapes in Ae only need relative x,y
    elif len(loc) == 2:
        bx = loc[0]
        by = 0
        bz = loc[1]

    else:
        print(f"Location conversion dimension is: {len(loc)}")
        return False

    #TODO - 100 might be better because smaller shape layers load faster in After Effects, 
    # but 1000 is more accurate for roundtripping cameras.

    custom_scale = bpy.context.scene.blenderae.export_scale_factor
    # print(f"custom_scale: {custom_scale}")
    scale_factor = 100.0 * custom_scale

    if bpy.context.scene.blenderae.center_offset and center_offset:
        x = (bx * scale_factor) / aspect + (width / 2.0)  # calculate AE's X position
        y = (-bz * scale_factor) + (height / 2.0)  # calculate AE's Y position

    else:
        x = (bx * scale_factor) / aspect
        y = -bz * scale_factor

    z = by * scale_factor  # calculate AE's Z position (0 for 2d face calculations)
    # print(f"loc_conv: {x}, {y}, {z}")
    return x, y, z


def rot_conv(rot):
    # Convert rotations
    rx = (
        degrees(rot[0]) - 90
    )  # AE's X orientation = blender's X rotation if 'ZYX' euler.
    ry = -degrees(
        rot[1]
    )  # AE's Y orientation is negative blender's Y rotation if 'ZYX' euler
    rz = -degrees(
        rot[2]
    )  # AE's Z orientation is negative blender's Z rotation if 'ZYX' euler
    return rx, ry, rz


def scale_conv(scale):
    # Blender scale is 1.0, needs to be 100% in AE
    # For a plane x,y scale are the same as Ae x,y (plane lies horizontal whereas in Ae it's vertical BUT x,y are the same 'relative' to the plane face).
    sx = scale[0] * 100.0
    sy = scale[1] * 100.0
    sz = scale[2] * 100.0
    return sx, sy, sz


# Not currently used, but may be useful in future
def select_only_cameras():
    curScene = bpy.context.scene
    for scene in bpy.data.scenes:
        bpy.context.window.scene = scene
        for ob in scene.objects:
            ob.select_set(ob.type == "CAMERA")

            # set scene camera to active for each viewlayer
            if scene.camera:
                for vl in scene.view_layers:
                    vl.objects.active = scene.camera
        bpy.context.window.scene = curScene


# ------------
# check planar
# ------------


def check_planar(obj, eps=0.0001):
    """check if an object is planar (flat)"""
    plane_defined_by = {"P0": "not found", "P1": "not found", "P2": "not found"}
    try:  # objects with no suitable elements are not considered
        verts = obj.data.vertices
        edges = obj.data.edges
        len_v = len(verts)
        len_e = len(edges)
    except Exception as e:
        print(f"check_planar error: {e}")
        return False

    if len_v == 3:
        # triangles are always planar
        return True

    if len_v != len_e:
        # number of verts unequal to number of edges
        return False

    # len_v == len_e:
    results = []
    np_points = []
    # verts created in the try-block
    np_points = [np.array(v.co) for v in verts if len(verts) > 3]
    if np_points:
        p0, p1, p2 = np_points[:3]
        va = p0 - p1
        vb = p0 - p2
        plane_defined_by["P0"] = [el for el in p0]
        plane_defined_by["P1"] = [el for el in p1]
        plane_defined_by["P2"] = [el for el in p2]

        to_check = len(np_points[3:])
        for i, n in enumerate(np_points[3:]):
            b = np.array([va, vb, p0 - n])
            # print(i, 'det is ', np.linalg.det(b))
            results.append(np.linalg.det(b))
            if abs(np.linalg.det(b)) > eps:
                # at least one vert not in the defining plane (check 1)
                return False

        if i == (to_check - 1):
            # print(str(plane_defined_by))
            print("Object " + obj.name + " is planar")
            return True

        else:
            # at least one vert not in the defining plane (check 2)
            return False


# -----------------------
# Object face matrix data
# -----------------------


def faces_verts(ob):
    """Returns a face global location and rotation, and associated vertex locations"""
    curr_active = bpy.context.view_layer.objects.active  # store currently active object
    bpy.context.view_layer.objects.active = ob  # switch active to argument obj

    # Switch to object mode
    bpy.ops.object.mode_set(mode="OBJECT")

    # Create bmesh obj with evaluated modifiers
    dg = bpy.context.evaluated_depsgraph_get()
    bm = bmesh.new()
    bm.from_object(ob, dg)  # creating the mesh copy
    bm.select_flush(True)
    bm.transform(ob.matrix_world)  # apply the matrix world of the object to the bmesh
    bm.normal_update()

    v_global = []
    loc_global = []
    rot_global = []
    verts_2d = []

    # For a 'plane' object with a single face, select the face (probably a better way to do this?)
    if len(bm.faces) == 1:
        for face in bm.faces:
            face.select = True

    # Calculate normal transformations for the selected faces in the bmesh
    for face in bm.faces:
        if face.select:
            n = face.normal
            t = face.calc_tangent_edge_pair().normalized()
            bt = n.cross(t).normalized()

            M = mathutils.Matrix([t, bt, n]).transposed().to_4x4()
            M.translation = face.calc_center_median()

            loc = list(M.translation)  # converted to list to remove 'vector'
            rot = list(
                M.to_euler("ZYX")
            )  # converted to list to remove 'vector', ZYX for Ae's rotation space

            x, y, z = loc_conv(
                loc
            )  # convert to Ae units, loc_conv returns a tuple of x,y,z.
            rx, ry, rz = rot_conv(rot)

            loc_global.append([x, y, z])
            rot_global.append([rx, ry, rz])

            # TODO Check verts are planar - pass through planar function
            # verts
            verts = [loc_conv(list(vert.co)) for vert in face.verts]
            v_global.append(verts)

            # Need each group of verts per face to be it's own list so store in temp list then add to verts_2d
            v_2d_temp = []
            for vert in face.verts:
                v_3d = M.inverted() @ vert.co
                v_2d = list(v_3d.to_2d())  # to_2d() requires vector input

                x, y, z = loc_conv(v_2d)
                v_2d_temp.append([x, y])
            verts_2d.append(v_2d_temp)

            # WARNING - below crashes Blender for unknown reason!
            # v_2d = [(inverted @ vert.co).to_2d() for vert in face.verts]
            # verts_2d.append(list(v_2d))

    bpy.context.view_layer.objects.active = curr_active  # reset active object
    curr_active.select_set(True)  # select object
    bm.free()  # free and prevent further access to bmesh

    # print("face loc_global: ", loc_global)
    return loc_global, rot_global, v_global, verts_2d


# ---------------------------------
# Get vertex locations and rotation
# ---------------------------------


def verts_only(obj):

    bpy.ops.object.mode_set(mode="OBJECT")

    # Create a bmesh of the object
    bm = create_bmesh_from_object(obj)  # create an empty BMesh

    # Get the object matrix_world data
    mw = obj.matrix_world
    # v_global = []
    # Get location for the selected verts in the bmesh
    v_global = [loc_conv(list(mw @ v.co)) for v in bm.verts if v.select]

    # for v in bm.verts:
    #     if v.select:
    #         v_global.append(
    #             loc_conv(list(mw @ v.co))
    #         )

    # TODO - include the vertex normal rotation
    # rot = list(vert.normal.to_track_quat('Z', 'Y').to_matrix().to_4x4())
    # https://blender.stackexchange.com/questions/167477/how-do-i-get-the-normal-vector-of-a-selected-vertexbmesh

    bm.free()  # free and prevent further access to bmesh

    return v_global


# --------------------------------
# Check if the array values change
# --------------------------------


def check_animated_properties(store_data):
    """Check whether a set of nested lists are the same, returns True if the values are identical"""

    def check_array(list_to_check):
        np_array = np.array(list_to_check)

        # Check array is greater than 1
        if np_array.size < 1:
            return False

        # If one dimension check all values against first value
        if np_array.ndim == 1:
            return np.all(np_array == np_array[0])

        trans_arr = np.array(np_array).T  # Transposing
        return all(
            np.allclose(trans_arr[i], trans_arr[i][0])
            for i in range(trans_arr.shape[0])
        )


    for value in store_data.values():
        # Only want to check the keys that are dicts
        try:
            if isinstance(value, dict):
                for object_key, object_values in value.items():
                    # print(f"Checking '{object_key}' data reduction...")
                    try:
                        for prop in object_values:
                            # print(prop)
                            if isinstance(object_values[prop], list) and check_array(
                                object_values[prop]
                            ):
                                object_values[prop] = [object_values[prop][0]]
                                print(f"{object_key}: {prop} reduced")

                            # else if the value is a dict, check the values of the dict
                            elif isinstance(object_values[prop], dict):
                                for sub_prop in object_values[prop]:
                                    if isinstance(
                                        object_values[prop][sub_prop], list
                                    ) and check_array(object_values[prop][sub_prop]):
                                        object_values[prop][sub_prop] = [
                                            object_values[prop][sub_prop][0]
                                        ]
                                        print(f"{object_key}: {prop} > {sub_prop} reduced")
                                    else:
                                        continue
                    except Exception as e:
                        print(f"'{object_key}' data reduction error: {e}")
                        continue
            else:
                continue

        except Exception as e:
            print(f"Error: {e}")
            continue

    return True


# --------------------
# Check for shape keys
# --------------------


def check_shape_keys(obj):
    if obj.data.shape_keys:
        print("Shape Keys: True")
        return True
    else:
        print("Shape Keys: False")
        return False


def user_data_dir():
    """
    Get OS specific data directory path for BlenderAe.
    :return: full path to the user-specific data dir
    """

    # Typical user data directories are:
    # macOS:    ~/Library/Application Support/BlenderAe
    # Unix:     ~/.local/share/BlenderAe   # or in $XDG_DATA_HOME, if defined
    # Win 10:   C:\Users\<username>\AppData\Local\BlenderAe
    # For Unix, we follow the XDG spec and support $XDG_DATA_HOME if defined.

    # Extendscript Folder.userData points to:
    # Win - C:\Users\[username]\Appdata\Roaming and,
    # Mac OS - ~/Library/Application Support (on Mac)

    # # get os specific path
    # if sys.platform.startswith("win"):
    #     os_path = os.getenv("APPDATA")
    # elif sys.platform.startswith("darwin"):
    #     os_path = "~/Library/Application Support"
    # else:
    #     # linux
    #     os_path = os.getenv("XDG_DATA_HOME", "~/.local/share")

    # return os.path.expanduser(os_path)

    # Get the current operating system
    operating_system = platform.system()

    # Create a Path object for the user's home directory
    home_dir = Path.home()

    # If the operating system is Windows, return the appdata roaming path
    if operating_system == "Windows":
        return home_dir.joinpath("AppData", "Roaming")

    # If the operating system is MacOS, return the appdata roaming path
    elif operating_system == "Darwin":
        return home_dir.joinpath("Library", "Application Support")

    # If the operating system is not recognized, return None
    else:
        return None


# ---------
# json dump
# ---------


def js_dump(data):
    # directory_path = os.path.dirname(getsourcefile(lambda: 0))

    # Pathlib doesn't work in VSCode and Blender using the blender extension? Check this out:
    directory_path = Path(__file__).parent
    data_folder = directory_path / "data"

    # Check to see if the user has access permissions to the directory to write
    if not os.access(directory_path, os.W_OK):
        print(f"No write access for the directory folder: {directory_path}")
        # change the directory_path to the user's data directory instead
        directory_path = user_data_dir() / "BlenderAe"
        data_folder = directory_path / "data"

    # create the data folder if it doesn't exist
    if not data_folder.exists():
        data_folder.mkdir(parents=True)

    full_path = data_folder / "BlenderAe_data.json"
    # full_path = os.path.join(directory_path, "data", "BlenderAe_data.json")
    # print("json path: ", full_path)

    try:
        with open(full_path, "w", encoding="utf-8") as fout:
            json.dump(data, fout, indent=2)

            # Check that the json file was written
            if full_path.exists():
                print(f"BlenderAe_data.json written to {full_path}")
            else:
                print(f"Error. BlenderAe_data.json file doesn't exist at {full_path}")

            return full_path

    except Exception as e:
        print(f"json writing error: {e}")
        return full_path


# --------------------
# Ae Application Path
# --------------------
def find_ae_path():

    # Possibilities:
    # 1. Ae only instance - ideal
    # 2. Pr only instance
    # 3. Pr and Ae open with Ae only instance (can't code differentiate between this and 2.)
    # 4. Pr and Ae (or background render) instance (>1)

    # Check if the file BlenderAe_Ae_path.txt exists in the Blender Foundation folder in the user data folder
    # If it does, read the file and return the path
    def get_ae_path_from_file():
        print("Checking for BlenderAe_Ae_path.txt file...")
        directory = user_data_dir()
        # directory = os.path.dirname(getsourcefile(lambda: 0))

        txtPath = Path(directory) / "Blender Foundation" / "BlenderAe_Ae_path.txt"

        # check if the file exists
        if not txtPath.exists():
            print(f"File not found: {txtPath}")
            return None

        print(f"Found: {txtPath}")

        with open(txtPath, "r") as f:
            AePath = Path(f.read())
            print(f"Ae path from file: {AePath}")
            # check if the path is valid
            if AePath.exists():
                print("AePath is valid")
                return AePath

            else:
                print("AePath is invalid")
                return None

    def get_ae_path_psutil(ae_processes):
        # Find the application path to After Effects using psutil
        # print("Checking Ae psutil processes...")

        try:
            for proc in psutil.process_iter():
                try:
                    # Check it's not a System Idle Process
                    if proc.pid == 0 or proc.exe() == "":
                        continue

                    # Check if the process name is in the list of After Effects processes
                    elif any(ae_proc in proc.name() for ae_proc in ae_processes):
                        ae_psutil_path = Path(proc.exe())
                        # Check that the ae_psutil_path is a valid application file
                        if ae_psutil_path.exists():
                            if platform.system() == "Windows":
                                if proc.name().endswith(".exe"):
                                    print(f"AePath from psutil: {ae_psutil_path}")
                                    return ae_psutil_path
                            elif platform.system() == "Darwin":
                                ae_psutil_darwin_path = (
                                    ae_psutil_path.parent.parent.parent
                                )
                                print(f"AePath from psutil: {ae_psutil_darwin_path}")
                                if Path(ae_psutil_darwin_path).exists():
                                    return ae_psutil_darwin_path
                            continue

                except (
                    psutil.NoSuchProcess,
                    psutil.AccessDenied,
                    psutil.ZombieProcess,
                ) as e:
                    print(f"psutil error: {e}")
                    return None

            return None

        except Exception as e:
            print(f"psutil process_iter error: {e}")
            return None

    def find_ae_windows_powershell():
        def powershell_run(cmd):
            return subprocess.run(
                ["powershell", "-Command", cmd],
                capture_output=True,
                text=True,
                universal_newlines=True,
            )

        # create a variable cmd for the powershell command to find AfterFX.exe or AfterFX (Beta).exe
        cmd = "Get-Process -Name 'AfterFX*' | Select-Object -ExpandProperty Path"

        print("Running PS command...")
        find_path = powershell_run(cmd)

        if find_path.stderr or find_path.returncode != 0:
            print(f"PowerShell error:\n {find_path.stderr}")
            print(f"Return code: {find_path.returncode}")
            app_path = False

        elif find_path.stdout and find_path.stdout != "":
            app_path = find_path.stdout.splitlines()
            app_path = app_path[0].strip()
            print(f"Path found: {app_path}")

        else:
            app_path = False

        return app_path

    def find_ae_windows_wmic():
        # wmic is deprecated in Windows 10 and 11
        try:
            print("\nRunning wmic command...")
            cmd = "wmic process where \"name='AfterFX.exe'\" get ExecutablePath"
            find_path = subprocess.run(
                cmd, capture_output=True, text=True, universal_newlines=True
            )

            # No Ae Path detected
            if find_path.stderr:
                print("find_Ae_windows() error: ", find_path.stderr)
                app_path = False

            elif find_path.stdout and find_path.stdout != "":
                app_path = find_path.stdout.splitlines()
                app_path = app_path[0].strip()
                print(f"Path found: {app_path}")

            else:
                app_path = False

            return app_path

        except Exception as e:
            print("wmic error: ", e)
            return False

    def find_ae_mac_applescript():
        mac_cmd = """
        tell application "System Events" to ¬
            set appAeList to the ¬
                name of every application process ¬
                whose background only is false ¬
                and name contains "After Effects"

        if appAeList is not {} then ¬
            return POSIX path of ¬
                (path to application ¬
                    (first item of appAeList))
        """
        print("Running find_Ae_Mac applescript...")
        mac_ae_path = subprocess.run(
            ["osascript", "-e", mac_cmd], capture_output=True, text=True
        )

        # Path was found
        if mac_ae_path.stdout and mac_ae_path.stdout != "":
            AePath = mac_ae_path.stdout.strip()
            print("Path found: ", AePath)

        # Path is empty
        elif mac_ae_path.stdout == "":
            print("Mac Ae not detected. Return code: ", mac_ae_path.returncode)
            AePath = False

        # Catch other cases
        else:
            print(
                "Error running Mac subprocess. mac_ae_path.stderr: ", mac_ae_path.stderr
            )
            AePath = False

        return AePath

    # -------------------------------
    ae_processes = [
        "AfterFX",
        "AfterFX (Beta)",
        "After Effects",
        "After Effects (Beta)",
    ]

    # Check the custom path first
    custom_AePath = bpy.context.scene.blenderae.ae_filepath
    # print("Custom Ae path: ", custom_AePath)
    if any(ae_proc in custom_AePath for ae_proc in ae_processes):
        AePath = custom_AePath
        connected_text = "Connected"
        ic = "LINKED"
        print("custom path, skipping detection: ", custom_AePath)
        return (AePath, connected_text, ic)

    # check txt file path
    AePath = str(get_ae_path_from_file())
    if AePath is not None and any(ae_proc in AePath for ae_proc in ae_processes):
        connected_text = "Connected"
        ic = "LINKED"
        # print(f"path type: {type(AePath)}")
        print("Text file path, skipping detection: ", AePath)
        return (AePath, connected_text, ic)

    # Before checking the platform specific options try psutil
    AePath = get_ae_path_psutil(ae_processes)
    if AePath is not None:
        connected_text = "Connected"
        ic = "LINKED"
        return (str(AePath), connected_text, ic)

    # -------------------------------
    # Platform specific detection
    # -------------------------------
    if sys.platform == "win32":
        # Otherwise try powershell and wmic
        try:
            AePath = find_ae_windows_powershell()

            if AePath and any(ae_proc in AePath for ae_proc in ae_processes):
                connected_text = "Connected"
                ic = "LINKED"

            else:
                AePath = ""
                connected_text = "----"
                ic = "UNLINKED"

            return (AePath, connected_text, ic)

        except Exception:
            print("PowerShell failed... trying wmic")
            try:
                AePath = find_ae_windows_wmic()

            except Exception:
                print("wmic failed... cancelling detection")
                AePath = ""
                connected_text = "----"
                ic = "UNLINKED"
                return (AePath, connected_text, ic)

            # AePath found, continue...
            if AePath and any(ae_proc in AePath for ae_proc in ae_processes):
                connected_text = "Connected"
                ic = "LINKED"

            # No active instance detected.
            else:
                AePath = ""
                connected_text = "----"
                ic = "UNLINKED"

            return (AePath, connected_text, ic)

    # ----------------
    elif sys.platform == "darwin":

        try:
            AePath = find_ae_mac_applescript()

            if AePath and any(ae_proc in AePath for ae_proc in ae_processes):
                connected_text = "Connected"
                ic = "LINKED"

            # No custom path and no active instance detected.
            else:
                AePath = ""
                connected_text = "----"
                ic = "UNLINKED"

            return (AePath, connected_text, ic)

        except Exception:
            print("find_ae_mac_applescript() error, skipping detection.")
            AePath = ""
            connected_text = "----"
            ic = "UNLINKED"

            return (AePath, connected_text, ic)

    else:
        print("System not detected as win32 or MacOS(darwin)!")
        AePath = ""
        connected_text = "----"
        ic = "UNLINKED"
        return (AePath, connected_text, ic)


# ---------------------
# Send to After Effects
# ---------------------
def ae_help_script_path():
    # Get the script path and add 'BlenderAe_Import.jsx' if it's in the same folder.
    # scriptPath = os.path.join(os.path.dirname(getsourcefile(lambda:0)), "BlenderAe_Help.jsxbin")
    directory = Path(__file__).parent.resolve()
    scriptPath = directory / "BlenderAe_Help.jsxbin"

    if Path(scriptPath).is_file() == False:
        print("jsxbin found: ", False)

        # scriptPath = os.path.join(os.path.dirname(getsourcefile(lambda:0)), "BlenderAe_Help.jsx")
        scriptPath = scriptPath = directory / "BlenderAe_Help.jsx"
        print("jsx found: ", Path(scriptPath).is_file())

        # if not Path(scriptPath).is_file():
        if Path(scriptPath).is_file() == False:
            print("Can't find the BlenderAe_Help.jsx or jsxbin file")
            scriptPath = ""
            return scriptPath
    else:
        print("Ae script path: ", scriptPath)

    return scriptPath
    # bpy.utils.user_resource('SCRIPTS', "addons") #The blender addon's folder path if needed


def ae_script_path():
    # Get the script path and add 'BlenderAe_Import.jsx' if it's in the same folder.
    # scriptPath = os.path.join(os.path.dirname(getsourcefile(lambda:0)), "BlenderAe_Import.jsxbin")
    directory = Path(__file__).parent.resolve()
    scriptPath = directory / "BlenderAe_Import.jsxbin"

    if Path(scriptPath).is_file() == False:
        print("jsxbin found: ", False)

        # scriptPath = os.path.join(os.path.dirname(getsourcefile(lambda:0)), "BlenderAe_Import.jsx")
        scriptPath = directory / "BlenderAe_Import.jsx"
        print("jsx found: ", Path(scriptPath).is_file())

        # if not Path(scriptPath).is_file():
        if Path(scriptPath).is_file() == False:
            print("Can't find the BlenderAe_Import.jsx or jsxbin file")
            scriptPath = ""
            return scriptPath
    else:
        print("Ae script path: ", scriptPath)

    return scriptPath
    # bpy.utils.user_resource('SCRIPTS', "addons") #The blender addon's folder path if needed


def copy_script():
    # set destination path
    # docu_dir = os.path.join(os.path.expanduser('~'),'Documents','BlenderAe')
    docu_dir = Path.home() / "Documents" / "BlenderAe"
    data_dir = docu_dir / "data"

    # if not os.path.isdir(docu_dir):
    if not Path(docu_dir).is_dir() or not Path(data_dir).is_dir():
        # os.mkdir(docu_dir)
        Path(docu_dir).mkdir(parents=True, exist_ok=True)
        Path(data_dir).mkdir(parents=True, exist_ok=True)

    # Check if the file exists in the destination folder, otherwise copy it (actually don't check, just copy and overwrite).
    # import_script_path = os.path.join(docu_dir, "BlenderAe_Import.jsxbin")

    directory = Path(__file__).parent.resolve()
    # Try to copy the script file
    # script_path = os.path.join(os.path.dirname(getsourcefile(lambda:0)), "BlenderAe_Import.jsxbin")
    script_path = directory / "BlenderAe_Import.jsxbin"
    if Path(script_path).is_file() == False:
        script_path = directory / "BlenderAe_Import.jsx"

        if Path(script_path).is_file() == False:
            print("Can't find the BlenderAe_Import.jsx or jsxbin file. Copy failed.")
            return False

    # Try to copy the json file
    json_path = directory / "data" / "BlenderAe_data.json"
    if Path(json_path).is_file() == False:
        print("Can't find the BlenderAe_Import.jsx or jsxbin file. Copy failed.")
        return False

    try:
        shutil.copy(script_path, docu_dir)
        shutil.copy(json_path, data_dir)
        print(f"Data file and script copied to '{docu_dir}'")
        return True

    # If source and destination are same
    except shutil.SameFileError:
        print("Source and destination represents the same file.")
        return False

    # If there is any permission issue
    except PermissionError:
        print("Permission denied.")
        return False

    # For other errors
    except Exception as e:
        print("copy_script() error")
        print(e)
        return False


def send_to_Ae():
    custom_AePath = bpy.context.scene.blenderae.ae_filepath
    ae_script = str(ae_script_path())

    # print(f"ae_script path: {ae_script}")

    # Windows
    if sys.platform == "win32":

        if ae_script != "" and custom_AePath != "":
            cmd = custom_AePath + " -r " + ae_script

            try:
                send_data = subprocess.run(
                    cmd, text=True, universal_newlines=True, timeout=5
                )

                if send_data.stderr:
                    print(f"Error sending jsx script to Ae: {send_data.stderr}")
                    return False
                else:
                    print(f"Script sent to Ae")
                    return True

            except subprocess.TimeoutExpired:
                print("Ae send data timeout")
                return False

        # If no path is set copy the import script to the Documents\Blender_Ae folder
        elif custom_AePath == "":
            return copy_script()

        else:
            print("Failed - win32 ae_script_path() or find_ae_path() was False")
            return False

    # Mac
    elif sys.platform == "darwin":
        if custom_AePath != "" and ae_script != "":
            cmd = f"""
                tell application "{custom_AePath}"
                    DoScriptFile "{ae_script}"
                end tell
                """
            send_data = subprocess.run(
                ["osascript", "-e", cmd], capture_output=True, text=True
            )

            # print("stderr: ", send_data.stderr)
            # print("stdout: ", send_data.stdout)
            # print("return: ", send_data.returncode)

            if send_data.returncode == 0:
                print("Data sent to After Effects Mac")
                return True

            else:
                print("Error sending data to After Effects Mac")
                return False
                # If no path is set copy the import script to the Documents\Blender_Ae folder

        elif custom_AePath == "":
            return copy_script()

        else:
            print(
                "Failed - darwin ae_script_path() or find_ae_path() was False and non-empty field"
            )
            return False

    else:
        print("System not detected as win32 or MacOS(darwin)!")
        return False


def help_Ae():
    custom_AePath = bpy.context.scene.blenderae.ae_filepath
    print("custom_AePath: ", custom_AePath)
    help_script = str(ae_help_script_path())

    if help_script != "":

        # Windows
        if sys.platform == "win32":

            cmd = custom_AePath + " -r " + help_script
            # print ("cmd: ", cmd)

            try:
                send_data = subprocess.run(cmd, capture_output=True, timeout=3)

                if send_data.stderr:
                    print("Error opening help: ", send_data.stderr)
                    return False

                elif send_data.stdout:
                    print("Help opened in After Effects")
                    return True

            except subprocess.TimeoutExpired:
                print("help_Ae timeout")
                # Assuming here that process was still executed
                return True

        # Mac
        elif sys.platform == "darwin":

            cmd = f"""
                tell application "{custom_AePath}"
                    DoScriptFile "{help_script}"
                end tell
                """

            send_data = subprocess.run(
                ["osascript", "-e", cmd], capture_output=True, text=True
            )

            # print("stderr: ", send_data.stderr)
            # print("stdout: ", send_data.stdout)
            # print("return: ", send_data.returncode)

            if send_data.returncode == 0:
                print("Data sent to After Effects Mac")
                return True

            else:
                print("Error sending data to After Effects Mac")
                return False

        else:
            print("System not detected as win32 or MacOS(darwin)!")
            return False

    else:
        print("Failed - ae_help_script_path() was empty")
        return False


def store_null_data(store_data, selection):

    for nulls in selection["nulls"]:
        null_name = nulls[1]
        null_obj = nulls[0]

        # Convert and unpack Null transformations for each object (properties at obj[0]... obj[1] is names)
        x, y, z, rx, ry, rz, sx, sy, sz = Ae_transforms(null_obj)

        # Add transformation values to Null dictionary
        store_data["nulls"][null_name]["position"].append([x, y, z])
        store_data["nulls"][null_name]["rotation"].append([rx, ry, rz])
        store_data["nulls"][null_name]["scale"].append([sx, sy, sz])


def store_cam_data(store_data, selection):
    def convert_lens(camera):
        if camera.data.sensor_fit == "HORIZONTAL":
            sensor = camera.data.sensor_width
            dimension = width

        elif camera.data.sensor_fit == "VERTICAL":
            sensor = camera.data.sensor_height
            dimension = height

        else:
            # Sensor width is used for AUTO sensor_fit setting
            sensor = camera.data.sensor_width
            # Check the width and height of the scene to determine the dimension
            dimension = width if width >= height else height

        return camera.data.lens * dimension / sensor * aspect

    for cam in selection["cameras"]:
        cam_name = cam[1]
        cam_obj = cam[0]

        # get zoom value for frame
        zoom = convert_lens(cam_obj)

        # Convert and unpack transformations
        x, y, z, rx, ry, rz, sx, sy, sz = Ae_transforms(cam_obj)

        # TODO - Compensate for shift_x and shift_y to in camera data, note that it affects the angle of view.
        # shift_x = cam_obj.data.shift_x
        # shift_y = cam_obj.data.shift_y

        # Add transformation values to dictionary
        store_data["cameras"][cam_name]["zoom"].append(zoom)
        store_data["cameras"][cam_name]["position"].append([x, y, z])
        store_data["cameras"][cam_name]["rotation"].append([rx, ry, rz])


def check_cam_markers():
    markers = currScene.timeline_markers

    if not markers or len(markers) <= 1:
        return False

    sorted_markers = [
        [marker.frame, clean_name(marker.camera)] for marker in markers if marker.camera
    ]
    sorted_markers = sorted(sorted_markers)

    for index, i in enumerate(sorted_markers):
        if (
            index == len(sorted_markers) - 1
        ):  # If the marker is the last marker make the end frame equal to the scene duration.
            marker_end = currScene.frame_end
        else:
            marker_end = sorted_markers[index + 1][
                0
            ]  # The marker end frame is 1 frame before the next camera marker start frame
        i.append(marker_end)  # append the end frame to the current marker

    # Check if any of the markers are outside of the timeline range. If they are then set them to the timeline range.
    for markers in sorted_markers:
        if markers[0] < currScene.frame_start:
            markers[0] = currScene.frame_start
        if markers[2] > currScene.frame_end:
            markers[2] = currScene.frame_end

    return sorted_markers


def store_lights_data(store_data, selection):

    for light in selection["lights"]:
        light_name = light[1]
        light_obj = light[0]

        light_type = light_obj.data.type

        # Spot lights in Ae have a cone angle and feather
        if light_type == "SPOT":
            cone_angle = degrees(light_obj.data.spot_size)
            cone_feather = light_obj.data.spot_blend * 100.0
            store_data["lights"][light_name]["cone_angle"].append(cone_angle)
            store_data["lights"][light_name]["cone_feather"].append(cone_feather)

        # #'PARALLEL' light in Ae - No option for cone_angle on a Parallel light in Ae! It's actually a 'Point of Interest'.
        # elif type == 'SUN':
        #     cone_angle = degrees(light_obj.data.angle)
        #     store_data['lights'][name_ae]['cone_angle'].append(cone_angle)

        # Light defaults are Point = 10w, Sun = 1, Spot = 10w, Area = 10w
        # specific light code for reference... https://docs.blender.org/api/current/bpy.types.Light.html?highlight=light%20types#bpy.types.Light.type

        if light_type == "SUN":
            energy = (light_obj.data.energy) * 100

        else:
            energy = light_obj.data.energy

        colour = (
            light_obj.data.color[0],
            light_obj.data.color[1],
            light_obj.data.color[2],
        )
        store_data["lights"][light_name]["energy"].append(energy)
        store_data["lights"][light_name]["colour"].append(colour)

        # Convert and unpack transformations for each object (properties at light_obj... obj[1] is names)
        x, y, z, rx, ry, rz, sx, sy, sz = Ae_transforms(light_obj)

        # Add transformation values to Null dictionary
        store_data["lights"][light_name]["position"].append([x, y, z])
        store_data["lights"][light_name]["rotation"].append([rx, ry, rz])


def store_vert_data(store_data, selection):
    # ---------------------------
    # Create verts_only dictionary
    for obj_and_name in selection["verts"]:

        name_ae = obj_and_name[1]
        obj = obj_and_name[0]

        # Convert and unpack transformations for each object (properties at [i][0]... [i][1] is names)
        x, y, z, rx, ry, rz, sx, sy, sz = Ae_transforms(obj)

        # Add transformation values to dictionary
        store_data["verts_only"][name_ae]["position"].append([x, y, z])
        store_data["verts_only"][name_ae]["rotation"].append([rx, ry, rz])
        store_data["verts_only"][name_ae]["scale"].append([sx, sy, sz])

        # Unpack verts_only for the selection return values
        v_global = verts_only(obj)

        # Add vert_locations to dictionary
        store_data["verts_only"][name_ae]["vert_locations"].append(v_global)


def check_vert_shape_keys(store_data, selection):

    for obj_and_name in selection["verts"]:

        name_ae = obj_and_name[1]
        obj = obj_and_name[0]

        # Check if there is shape keys (set to true of false)
        store_data["verts_only"][name_ae]["shape_keys"] = check_shape_keys(obj)

        # If no shape keys and scale doesn't begin at 0 on the first frame (ie. all the verts are in the same location at the beginning), then reduce keyframes.
        if not store_data["verts_only"][name_ae]["shape_keys"] and store_data[
            "verts_only"
        ][name_ae]["scale"][0] != [0, 0, 0]:
            store_data["verts_only"][name_ae]["vert_locations"] = [
                store_data["verts_only"][name_ae]["vert_locations"][0]
            ]


def create_bmesh_from_object(obj, evaluated=True, transform=False, triangulate=False):
    """
    Returns an evaluated, transformed, triangulated copy of the mesh
    """
    assert obj.type == "MESH"

    me = obj.data
    if not evaluated and obj.mode == "EDIT":
        bm = bmesh.new()
        # bm = bmesh.from_edit_mesh(me).copy()
        bm_orig = bmesh.from_edit_mesh(me)
        bm = bm_orig.copy()

    elif not evaluated:
        bm = bmesh.new()
        bm.from_mesh(me)

    else:
        # Switch to object mode
        bpy.ops.object.mode_set(mode="OBJECT")

        # Create bmesh obj with evaluated modifiers
        dg = bpy.context.evaluated_depsgraph_get()
        bm = bmesh.new()
        bm.from_object(obj, dg)  # creating the mesh copy
        bm.normal_update()

    bm.select_flush(True)

    if transform:
        bm.transform(obj.matrix_world)

    if triangulate:
        bmesh.ops.triangulate(bm, faces=bm.faces)

    return bm


def bmesh_face_data(obj, bm):

    curr_active = bpy.context.view_layer.objects.active  # store currently active object
    bpy.context.view_layer.objects.active = obj  # switch active to argument obj

    # For a 'plane' object with a single face, select the face (probably a better way to do this?)
    if len(bm.faces) == 1:
        for face in bm.faces:
            face.select = True

    v_global = []
    loc_global = []
    rot_global = []
    verts_2d = []

    # print ("Face length: ", len(bm.faces))

    # Calculate normal transformations for the selected faces in the bmesh
    for face in bm.faces:
        if face.select:
            n = face.normal
            t = face.calc_tangent_edge_pair().normalized()
            bt = n.cross(t).normalized()

            M = mathutils.Matrix([t, bt, n]).transposed().to_4x4()
            M.translation = face.calc_center_median()

            loc = list(M.translation)
            rot = list(M.to_euler("ZYX"))

            x, y, z = loc_conv(loc)
            rx, ry, rz = rot_conv(rot)

            loc_global.append([x, y, z])
            rot_global.append([rx, ry, rz])

            # TODO Check verts are planar - pass through planar function
            # verts
            verts = [loc_conv(list(vert.co)) for vert in face.verts]
            v_global.append(verts)

            # Need each group of verts per face to be it's own list. Store in temp list then add to verts_2d.
            v_2d_temp = []
            for vert in face.verts:
                v_3d = M.inverted() @ vert.co
                v_2d = list(v_3d.to_2d())  # to_2d() requires vector input

                x, y, z = loc_conv(v_2d, False)
                v_2d_temp.append([x, y])

            verts_2d.append(v_2d_temp)

    bpy.context.view_layer.objects.active = curr_active  # reset active object
    curr_active.select_set(True)
    bm.free()

    return loc_global, rot_global, v_global, verts_2d


def store_face_obj_transform_data(store_data, obj, obj_name):

    # Convert and unpack transformations for each object
    x, y, z, rx, ry, rz, sx, sy, sz = Ae_transforms(obj)

    # Add transformation values to dictionary
    store_data["faces_simple"][obj_name]["position"].append([x, y, z])
    store_data["faces_simple"][obj_name]["rotation"].append([rx, ry, rz])
    store_data["faces_simple"][obj_name]["scale"].append([sx, sy, sz])

    return


def store_bone_data(store_data, selection):
    
    def create_empty(name: str, size=1, display_type="ARROWS", rotation_mode='XYZ'):
        """create an empty with a name and type"""

        empty_obj = bpy.data.objects.new(name, None)
        empty_obj.empty_display_size = size

        # enum in ['PLAIN_AXES', 'ARROWS', 'SINGLE_ARROW', 'CIRCLE', 'CUBE', 'SPHERE', 'CONE', 'IMAGE']
        empty_obj.empty_display_type = display_type

        # ZYX matches Ae rotation space
        empty_obj.rotation_mode = rotation_mode

        return empty_obj

    for armature, armature_name, bone in selection["bones"]:

        # pose_bone_matrix = armature.matrix_world @ bone.matrix

        edit_bone_rotation = rot_conv(armature.matrix_world.to_euler("ZYX"))
        bone_rotation = rot_conv(bone.matrix.to_euler("ZYX"))
        head_pos = loc_conv((armature.matrix_world @ bone.head), True)
        tail_pos = loc_conv((armature.matrix_world @ bone.tail), True)
        bone_length = np.linalg.norm(np.array(head_pos) - np.array(tail_pos))

        axis, bone_roll = bone.bone.AxisRollFromMatrix(bone.matrix.to_3x3())

        store_data["bones"][armature_name][bone.name]["head_position"].append(head_pos)

        if bpy.context.scene.blenderae.include_tails:
            store_data["bones"][armature_name][bone.name]["tail_position"].append(tail_pos)
        
        store_data["bones"][armature_name][bone.name]["bone_rotation"].append(bone_rotation)
        store_data["bones"][armature_name][bone.name]["bone_roll"].append(bone_roll)
        store_data["bones"][armature_name][bone.name]["bone_length"].append(bone_length)

    return


# ----------------------
# create data dictionary
# ----------------------


def create_object_dicts(selection):
    store_data = {
        "fileDir": get_blender_path(),
        # 'pathDir': user_data_dir(),
        "pathDir": os.path.dirname(getsourcefile(lambda: 0)),
        "width": width,
        "height": height,
        "aspect": aspect,
        "fps": fps,
        "first_frame": first_frame,
        "last_frame": last_frame,
        "duration": duration,
        "lights": {},
        "nulls": {},
        "cam_markers": [],
        "cameras": {},
        "faces_simple": {},
        "verts_only": {},
        "bones": {},
    }

    # Bones
    for armature, armature_name, bone in selection["bones"]:
        # print(f"armature: {armature}, armature_name: {armature_name}, bone: {bone}")

        store_data["bones"]["include_tails"] = bpy.context.scene.blenderae.include_tails

        # Create Armature's dictionary
        if armature_name not in store_data["bones"]:
            store_data["bones"][armature_name] = {}

        store_data["bones"][armature_name][bone.name] = {
            "head_position": [],
            "tail_position": [],
            "bone_rotation": [],
            "bone_roll": [],
            "bone_length": [],
        }

    # Cameras
    if check_cam_markers():
        # store_data['cam_markers'].append(check_cam_markers())
        store_data["cam_markers"] = check_cam_markers()

    for cam in selection["cameras"]:
        # Create Camera's dictionary
        name_ae = cam[1]
        # includes all camera names
        store_data["cameras"][name_ae] = {
            "zoom": [],
            "position": [],
            "rotation": [],
            "camera type": cam[0].data.type,
        }

    # Lights
    for light in selection["lights"]:
        name_ae = light[1]
        light_type = light[0].data.type

        store_data["lights"][name_ae] = {
            "type": light_type,
            "cone_angle": [],
            "cone_feather": [],
            "energy": [],
            "colour": [],
            "position": [],
            "rotation": [],
        }

    # Nulls
    for nulls in selection["nulls"]:
        name_ae = nulls[1]
        store_data["nulls"][name_ae] = {
            "position": [],
            "rotation": [],
            "scale": [],
        }

    # Faces
    for face in selection["faces"]:

        name_ae = face[1]
        obj = face[0]

        store_data["faces_simple"][name_ae] = {
            "shape_keys": False,
            "position": [],
            "rotation": [],
            "scale": [],
            "face_loc_global": [],
            "face_rot_global": [],
            "face_verts_global": [],
            "face_verts_2d": [],
        }

    # Verts_only
    for verts_and_name in selection["verts"]:

        name_ae = verts_and_name[1]
        obj = verts_and_name[0]

        store_data["verts_only"][name_ae] = {
            "shape_keys": False,
            "vert_locations": [],
            # 'vert_rotations': [],
            "position": [],
            "rotation": [],
            "scale": [],
        }

    return store_data


def get_objects_data(store_data, selection):
    def store_face_data(store_data, obj, obj_name):
        # create bmesh without transform
        bm = create_bmesh_from_object(
            obj, evaluated=True, transform=False, triangulate=False
        )

        # Unpack face data
        face_loc_global, face_rot_global, verts_global, verts_2d = bmesh_face_data(
            obj,
            bm,
        )

        bm.free()

        # Add face normal and verts
        store_data["faces_simple"][obj_name]["face_loc_global"].append(face_loc_global)
        store_data["faces_simple"][obj_name]["face_rot_global"].append(face_rot_global)
        store_data["faces_simple"][obj_name]["face_verts_global"].append(verts_global)
        store_data["faces_simple"][obj_name]["face_verts_2d"].append(verts_2d)

        return

    # For faces if no shape keys only need to get initial bmesh.
    # Loop through face objects and check shape keys before timeline loop
    for obj in selection["faces"]:
        obj_name = obj[1]
        obj = obj[0]

        store_data["faces_simple"][obj_name]["shape_keys"] = check_shape_keys(obj)

        if not store_data["faces_simple"][obj_name]["shape_keys"]:
            store_face_data(store_data, obj, obj_name)

        # Add initial untransformed bmesh location and rotation
        bm_location = loc_conv([0, 0, 0])
        bm_rotation = rot_conv([0, 0, 0])
        store_data["faces_simple"][obj_name]["bmesh_initial_position"] = bm_location
        store_data["faces_simple"][obj_name]["bmesh_initial_rotation"] = bm_rotation

    # Now loop through the timeline for other objects
    for f in range(currScene.frame_start, currScene.frame_end):
        currScene.frame_set(f)

        # store the data
        store_null_data(store_data, selection)
        store_cam_data(store_data, selection)
        store_lights_data(store_data, selection)
        store_bone_data(store_data, selection)

        # Needed to pull out this face loop to separate the shape key check earlier
        for obj in selection["faces"]:
            obj_name = obj[1]
            obj = obj[0]

            store_face_obj_transform_data(store_data, obj, obj_name)

            # If there is shape_keys store data for every frame
            if store_data["faces_simple"][obj_name]["shape_keys"]:
                store_face_data(store_data, obj, obj_name)

        store_vert_data(store_data, selection)

    check_vert_shape_keys(store_data, selection)
    # Check the animated values to see if they need reduction (if all the same values)
    # HACK - Currently getting all the values then simplifying using numpy
    # Pro's - this accounts for any driver's modifiers and parents
    # Con's - not exporting the actual blender keyframes

    check_animated_properties(store_data)
    js_dump(store_data)


# ---------
# Run
# ---------


def export_json_ae():

    print("\nRunning JSON Export...")

    global obs
    global currScene
    global aspect_x
    global aspect_y
    global width
    global height
    global aspect
    global fps
    global first_frame
    global last_frame
    global duration

    obs = bpy.context.selected_objects
    currScene = bpy.context.scene
    aspect_x = currScene.render.pixel_aspect_x
    aspect_y = currScene.render.pixel_aspect_y
    width = currScene.render.resolution_x
    height = currScene.render.resolution_y
    aspect = aspect_x / aspect_y
    fps = floor(currScene.render.fps / (currScene.render.fps_base) * 1000.0) / 1000.0
    first_frame = currScene.frame_start
    last_frame = currScene.frame_end
    duration = currScene.frame_end - currScene.frame_start

    # if the duration is 0 then warn the user and return false with an error message
    if duration == 0:
        return (False, "Duration is 0, please set the start and end frames")

    curr_frame = currScene.frame_current
    curr_mode = bpy.context.object.mode

    if len(obs) != 0:
        # get the selected objects
        selection = get_selected_objects(obs)
        store_data = create_object_dicts(selection)
        get_objects_data(store_data, selection)
        send_to_Ae()

        currScene.frame_set(curr_frame)
        bpy.ops.object.mode_set(mode=curr_mode)
        return (True, "Data exported to After Effects")

    else:
        return (False, "No objects selected")
