# ##### BEGIN GPL LICENSE BLOCK #####
#
# This file is part of the BlenderAe Addon.
#
# BlenderAe is a Blender Addon that allows exporting of Blender data for After Effects.
# Copyright (C) 2021 Daniel Peterson
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of  MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ##### END GPL LICENSE BLOCK #####

import platform
from pathlib import Path

import bpy
from bpy.types import Operator

from .BlenderAe_AeLayers_Import import ae_subprocess_execute, data_import
from .BlenderAe_Export import export_json_ae, find_ae_path, help_Ae


class BLENDERAE_OT_connect_ae(Operator):
    bl_idname = "export_scene.connect_ae"
    bl_label = "Connect to Ae"
    bl_description = "Connect to active instance of After Effects"

    def execute(self, context):
        custom_AePath = bpy.context.scene.blenderae.ae_filepath
        operating_system = platform.system()

        # Check psutil is installed
        try:
            import psutil
        except ImportError:
            self.report(
                {"ERROR"},
                "Please install the BlenderAe dependencies and restart Blender.",
            )
            bpy.ops.preferences.addon_show(module=__package__)
            # return {"CANCELLED"}

        # TODO - I can set these directly in the BlenderAe_Export.py file(?), leaving here for now.
        if custom_AePath == "":
            AePath, connected_text, ic = find_ae_path()
            bpy.context.scene.blenderae.ae_connected = connected_text
            bpy.context.scene.blenderae.ae_ic = ic
            bpy.context.scene.blenderae.ae_filepath = str(AePath)

            if bpy.context.scene.blenderae.ae_connected == "----":
                self.report({"ERROR"}, "No After Effects instance detected.")

        # If on windows system, check for AfterFX.exe or AfterFX (Beta).exe
        if operating_system == "Windows" and custom_AePath != "":
            if "AfterFX.exe" in custom_AePath or "AfterFX (Beta).exe" in custom_AePath:
                bpy.context.scene.blenderae.ae_connected = "Connected"
                bpy.context.scene.blenderae.ae_ic = "LINKED"
            else:
                self.set_to_unlinked()

        # If on Mac system, check for After Effects in the name of the base path
        if operating_system == "Darwin" and custom_AePath != "":
            path = Path(custom_AePath)
            if "After Effects" in path.name:
                bpy.context.scene.blenderae.ae_connected = "Connected"
                bpy.context.scene.blenderae.ae_ic = "LINKED"
            else:
                self.set_to_unlinked()
        return {"FINISHED"}

    def set_to_unlinked(self):
        bpy.context.scene.blenderae.ae_connected = "----"
        bpy.context.scene.blenderae.ae_ic = "UNLINKED"
        bpy.context.scene.blenderae.ae_filepath = ""

        self.report({"ERROR"}, "Incorrect path detected.")


class BLENDERAE_OT_disconnect_ae(Operator):
    bl_idname = "export_scene.disconnect_ae"
    bl_label = "Disconnect"
    bl_description = "Disconnect from active instance of After Effects"

    def execute(self, context):
        bpy.context.scene.blenderae.ae_filepath = ""
        bpy.context.scene.blenderae.ae_connected = "----"
        bpy.context.scene.blenderae.ae_ic = "UNLINKED"

        return {"FINISHED"}


class BLENDERAE_OT_export_ae(Operator):
    """Export data to json for After Effects"""

    bl_idname = "export_scene.export_ae"
    bl_label = "Export to Ae"
    bl_description = "Export selected to After Effects"

    @classmethod
    def poll(cls, context):
        obj = context.object

        if obj is not None:
            return obj.mode in ["OBJECT", "EDIT", "POSE"]

    def execute(self, context):
        exported = export_json_ae()

        if bpy.context.scene.blenderae.ae_connected == "Connected":
            if exported[0]:
                self.report({"INFO"}, exported[1])
            else:
                self.report({"ERROR"}, exported[1])

        else:
            print("Not connected, saving to docs")
            self.report(
                {"INFO"},
                f"Data exported to '{Path.home().joinpath('Documents', 'BlenderAe')}'",
            )

        return {"FINISHED"}


class BLENDERAE_OT_help_ae(Operator):
    bl_idname = "export_scene.help_ae"
    bl_label = "Help & License (opens in Ae)"
    bl_description = "Open Help in After Effects"

    def execute(self, context):
        if bpy.context.scene.blenderae.ae_connected == "Connected":
            if help_Ae():
                self.report({"INFO"}, "Help opened in After Effects Complete")
            else:
                self.report({"ERROR"}, "Help Error")

        else:
            print("not connected sorry")
            self.report({"ERROR"}, "No After Effects instance detected.")

        return {"FINISHED"}


class BLENDERAE_OT_import_ae_layers(Operator):
    """Import json data from After Effects"""

    bl_idname = "blenderae.import_ae_layers"
    bl_label = "Import Layers"
    bl_description = "Import selected After Effects layers. 3D Camera tracks, solids, nulls currently supported"

    def execute(self, context):
        # print(f"ae_connected: {bpy.context.scene.blenderae.ae_connected}")
        if bpy.context.scene.blenderae.ae_connected == "Connected":
            if ae_subprocess_execute():
                data_import()
                self.report({"INFO"}, "Imported After Effects layers")

            else:
                self.report({"ERROR"}, "Data import error.")

        else:
            self.report({"ERROR"}, "No After Effects instance connected.")

        return {"FINISHED"}
